from dissect.jffs.exceptions import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)
from dissect.jffs.jffs2 import JFFS2

__all__ = [
    "JFFS2",
    "Error",
    "FileNotFoundError",
    "NotADirectoryError",
    "NotASymlinkError",
]



from __future__ import annotations

import gzip
from pathlib import Path
from typing import TYPE_CHECKING, BinaryIO

import pytest

if TYPE_CHECKING:
    from collections.abc import Iterator


def absolute_path(filename: str) -> Path:
    return Path(__file__).parent / filename


def gzip_file(filename: str) -> Iterator[BinaryIO]:
    with gzip.GzipFile(absolute_path(filename), "rb") as fh:
        yield fh


@pytest.fixture
def jffs2_bin() -> Iterator[BinaryIO]:
    yield from gzip_file("data/jffs2.bin.gz")


@pytest.fixture
def jffs2_zlib() -> Iterator[BinaryIO]:
    yield from gzip_file("data/jffs2.zlib.bin.gz")


@pytest.fixture
def jffs2_router() -> Iterator[BinaryIO]:
    yield from gzip_file("data/router.bin.gz")



from __future__ import annotations

import hashlib
from datetime import datetime, timezone
from io import BytesIO
from typing import BinaryIO

import pytest

from dissect.jffs.exceptions import FileNotFoundError
from dissect.jffs.jffs2 import JFFS2


def test_jffs2_uncompressed(jffs2_bin: BinaryIO) -> None:
    fs = JFFS2(jffs2_bin)

    root = fs.root
    assert root.is_dir()
    assert root.nlink == 4  #  3 from root and 1 from subdirectory
    assert list(root.listdir().keys()) == ["foo", "test.txt"]

    test_file = fs.get("/test.txt")
    assert test_file.is_file()
    assert test_file.nlink == 1
    assert test_file.atime == datetime(2023, 6, 23, 20, 27, 20, tzinfo=timezone.utc)
    assert test_file.ctime == datetime(2023, 6, 23, 20, 27, 20, tzinfo=timezone.utc)
    assert test_file.mtime == datetime(2023, 6, 23, 20, 27, 20, tzinfo=timezone.utc)
    assert test_file.open().read() == b"contents\n"

    link_file = fs.get("/foo/bar/link.txt")
    assert link_file.is_symlink()
    assert link_file.nlink == 1
    assert link_file.link == "/test.txt"


def test_jffs2_zlib_compressed(jffs2_zlib: BinaryIO) -> None:
    fs = JFFS2(jffs2_zlib)

    root = fs.root
    assert root.is_dir()
    assert list(root.listdir().keys()) == ["folder", "fox-banner.png", "link.txt", "readme.md"]

    large_file = fs.get("/fox-banner.png")
    assert large_file.is_file()
    assert large_file.size == 103911

    for time in ["atime", "ctime", "mtime"]:
        assert getattr(large_file, time) == datetime(2023, 9, 29, 12, 9, 23, tzinfo=timezone.utc)

    with large_file.open() as fh:
        data = fh.read()

    assert len(data) == 103911
    assert hashlib.sha1(data).hexdigest() == "a0dc9f8d64cd34a96778984f3f13ca75cc99bbe4"


def test_jffs2_deleted_file(jffs2_zlib: BinaryIO) -> None:
    del_dirent = """
    851901e036000000a4e155df01000000010000000000000073be16650e080000
    99e595847c15af55666f782d62616e6e65722e706e67
    """
    data = jffs2_zlib.read() + bytes.fromhex(del_dirent)

    fs = JFFS2(BytesIO(data))
    assert len(fs._lost_found) == 1

    with pytest.raises(FileNotFoundError):
        fs.get("/fox-banner.png")

    lost_found = fs.get("/lost+found")
    assert list(lost_found.listdir().keys()) == ["fox-banner.png_ino_3_pino_1_ver_1"]

    deleted_file = fs.get("/lost+found/fox-banner.png_ino_3_pino_1_ver_1")
    assert deleted_file.parent is None

    with deleted_file.open() as fh:
        data = fh.read()

    assert len(data) == 103911
    assert hashlib.sha1(data).hexdigest() == "a0dc9f8d64cd34a96778984f3f13ca75cc99bbe4"


def test_jffs2_first_ctf_2023(jffs2_router: BinaryIO) -> None:
    fs = JFFS2(jffs2_router)

    assert len(fs._dirents) == 133
    assert len(fs._inodes) == 1384
    assert len(fs._lost_found) == 0

    file = fs.get("/upper/www/src/static/images/background.jpg")
    with file.open() as fh:
        data = fh.read()

    assert len(data) == 45469
    assert hashlib.sha1(data).hexdigest() == "669b00be651d29e618befddbfecf9f9ee82b93f9"


def test_jffs2_out_of_order_versions(jffs2_zlib: BinaryIO) -> None:
    old_dirent = """
    851901e03000000078be3efa010000000100000004000000b3be1665080a0000
    61748c85eecaede66c696e6b2e747874
    """
    data = jffs2_zlib.read() + bytes.fromhex(old_dirent)

    fs = JFFS2(BytesIO(data))
    assert fs._dirents[1][b"link.txt"][0].version == 1
    assert fs._dirents[1][b"link.txt"][1].version == 2



# References:
# - https://github.com/torvalds/linux/blob/master/include/uapi/linux/jffs2.h
# - https://github.com/torvalds/linux/blob/master/include/linux/fs_types.h

from __future__ import annotations

import stat

from dissect.cstruct import cstruct

jffs2_def = """
/* these are defined by POSIX and also present in glibc's dirent.h */
#define DT_UNKNOWN                      0
#define DT_FIFO                         1
#define DT_CHR                          2
#define DT_DIR                          4
#define DT_BLK                          6
#define DT_REG                          8
#define DT_LNK                          10
#define DT_SOCK                         12
#define DT_WHT                          14

/* Values we may expect to find in the 'magic' field */
#define JFFS2_OLD_MAGIC_BITMASK         0x1984
#define JFFS2_MAGIC_BITMASK             0x1985
#define KSAMTIB_CIGAM_2SFFJ             0x8519      /* For detecting wrong-endian fs */
#define JFFS2_EMPTY_BITMASK             0xffff
#define JFFS2_DIRTY_BITMASK             0x0000

/* Summary node MAGIC marker */
#define JFFS2_SUM_MAGIC                 0x02851885

/* We only allow a single char for length, and 0xFF is empty flash so
   we don't want it confused with a real length. Hence max 254.
*/
#define JFFS2_MAX_NAME_LEN              254

/* How small can we sensibly write nodes? */
#define JFFS2_MIN_DATA_LEN              128

#define JFFS2_COMPR_NONE                0x00
#define JFFS2_COMPR_ZERO                0x01        /* All data in such an inode should be 0x00 */
#define JFFS2_COMPR_RTIME               0x02        /* https://github.com/torvalds/linux/blob/master/fs/jffs2/compr_rtime.c */
#define JFFS2_COMPR_RUBINMIPS           0x03        /* Deprecated. */
#define JFFS2_COMPR_COPY                0x04        /* Never implemented. */
#define JFFS2_COMPR_DYNRUBIN            0x05        /* https://github.com/torvalds/linux/blob/master/fs/jffs2/compr_rubin.c */
#define JFFS2_COMPR_ZLIB                0x06
#define JFFS2_COMPR_LZO                 0x07

/* Compatibility flags. */
#define JFFS2_COMPAT_MASK               0xc000      /* What do to if an unknown nodetype is found */
#define JFFS2_NODE_ACCURATE             0x2000
/* INCOMPAT: Fail to mount the filesystem */
#define JFFS2_FEATURE_INCOMPAT          0xc000
/* ROCOMPAT: Mount read-only */
#define JFFS2_FEATURE_ROCOMPAT          0x8000
/* RWCOMPAT_COPY: Mount read/write, and copy the node when it's GC'd */
#define JFFS2_FEATURE_RWCOMPAT_COPY     0x4000
/* RWCOMPAT_DELETE: Mount read/write, and delete the node when it's GC'd */
#define JFFS2_FEATURE_RWCOMPAT_DELETE   0x0000

#define JFFS2_NODETYPE_DIRENT           (JFFS2_FEATURE_INCOMPAT | JFFS2_NODE_ACCURATE | 1)
#define JFFS2_NODETYPE_INODE            (JFFS2_FEATURE_INCOMPAT | JFFS2_NODE_ACCURATE | 2)
#define JFFS2_NODETYPE_CLEANMARKER      (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 3)
#define JFFS2_NODETYPE_PADDING          (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 4)

#define JFFS2_NODETYPE_SUMMARY          (JFFS2_FEATURE_RWCOMPAT_DELETE | JFFS2_NODE_ACCURATE | 6)

#define JFFS2_NODETYPE_XATTR            (JFFS2_FEATURE_INCOMPAT | JFFS2_NODE_ACCURATE | 8)
#define JFFS2_NODETYPE_XREF             (JFFS2_FEATURE_INCOMPAT | JFFS2_NODE_ACCURATE | 9)

/* XATTR Related */
#define JFFS2_XPREFIX_USER              1           /* for "user." */
#define JFFS2_XPREFIX_SECURITY          2           /* for "security." */
#define JFFS2_XPREFIX_ACL_ACCESS        3           /* for "system.posix_acl_access" */
#define JFFS2_XPREFIX_ACL_DEFAULT       4           /* for "system.posix_acl_default" */
#define JFFS2_XPREFIX_TRUSTED           5           /* for "trusted.*" */

#define JFFS2_ACL_VERSION               0x0001

#define JFFS2_INO_FLAG_PREREAD          1           /* Do read_inode() for this one at
                                                       mount time, don't wait for it to
                                                       happen later */
#define JFFS2_INO_FLAG_USERCOMPR        2           /* User has requested a specific
                                                       compression type */

struct jffs2_unknown_node {
    /* All start like this */
    uint16  magic;
    uint16  nodetype;
    uint32  totlen;             /* So we can skip over nodes we don't grok */
    uint32  hdr_crc;
};

struct jffs2_raw_dirent {
    uint16  magic;
    uint16  nodetype;           /* == JFFS2_NODETYPE_DIRENT */
    uint32  totlen;
    uint32  hdr_crc;
    uint32  pino;
    uint32  version;
    uint32  ino;                /* == zero for unlink */
    uint32  mctime;
    uint8   nsize;
    uint8   type;
    uint8   unused[2];
    uint32  node_crc;
    uint32  name_crc;
    char    name[nsize];
};

/* The JFFS2 raw inode structure: Used for storage on physical media.  */
/* The uid, gid, atime, mtime and ctime members could be longer, but
   are left like this for space efficiency. If and when people decide
   they really need them extended, it's simple enough to add support for
   a new type of raw node.
*/
struct jffs2_raw_inode {
    uint16  magic;              /* A constant magic number.  */
    uint16  nodetype;           /* == JFFS2_NODETYPE_INODE */
    uint32  totlen;             /* Total length of this node (inc data, etc.) */
    uint32  hdr_crc;
    uint32  ino;                /* Inode number.  */
    uint32  version;            /* Version number.  */
    uint32  mode;               /* The file's type or mode.  */
    uint16  uid;                /* The file's owner.  */
    uint16  gid;                /* The file's group.  */
    uint32  isize;              /* Total resultant size of this inode (used for truncations)  */
    uint32  atime;              /* Last access time.  */
    uint32  mtime;              /* Last modification time.  */
    uint32  ctime;              /* Change time.  */
    uint32  offset;             /* Where to begin to write.  */
    uint32  csize;              /* (Compressed) data size */
    uint32  dsize;              /* Size of the node's data. (after decompression) */
    uint8   compr;              /* Compression algorithm used */
    uint8   usercompr;          /* Compression algorithm requested by the user */
    uint16  flags;              /* See JFFS2_INO_FLAG_* */
    uint32  data_crc;           /* CRC for the (compressed) data.  */
    uint32  node_crc;           /* CRC for the raw inode (excluding data)  */
    // char    data[csize];
};

struct jffs2_raw_xattr {
    uint16  magic;
    uint16  nodetype;           /* = JFFS2_NODETYPE_XATTR */
    uint32  totlen;
    uint32  hdr_crc;
    uint32  xid;                /* XATTR identifier number */
    uint32  version;
    uint8   xprefix;
    uint8   name_len;
    uint16  value_len;
    uint32  data_crc;
    uint32  node_crc;
    uint8   data[0];
};

struct jffs2_raw_xref {
    uint16  magic;
    uint16  nodetype;           /* = JFFS2_NODETYPE_XREF */
    uint32  totlen;
    uint32  hdr_crc;
    uint32  ino;                /* inode number */
    uint32  xid;                /* XATTR identifier number */
    uint32  xseqno;             /* xref sequential number */
    uint32  node_crc;
};

struct jffs2_raw_summary {
    uint16  magic;
    uint16  nodetype;           /* = JFFS2_NODETYPE_SUMMARY */
    uint32  totlen;
    uint32  hdr_crc;
    uint32  sum_num;            /* number of sum entries*/
    uint32  cln_mkr;            /* clean marker size, 0 = no cleanmarker */
    uint32  padded;             /* sum of the size of padding nodes */
    uint32  sum_crc;            /* summary information crc */
    uint32  node_crc;           /* node crc */
    uint32  sum[0];             /* inode summary info */
};

/* Data payload for device nodes. */
union jffs2_device_node {
    uint16  old_id;
    uint32  new_id;
};
"""  # noqa

c_jffs2 = cstruct().load(jffs2_def)

JFFS2_MAGIC_NUMBERS = (c_jffs2.JFFS2_MAGIC_BITMASK, c_jffs2.JFFS2_OLD_MAGIC_BITMASK)

DT_MAP = {
    c_jffs2.DT_UNKNOWN: None,
    c_jffs2.DT_FIFO: stat.S_IFIFO,
    c_jffs2.DT_CHR: stat.S_IFCHR,
    c_jffs2.DT_DIR: stat.S_IFDIR,
    c_jffs2.DT_BLK: stat.S_IFBLK,
    c_jffs2.DT_REG: stat.S_IFREG,
    c_jffs2.DT_LNK: stat.S_IFLNK,
    c_jffs2.DT_SOCK: stat.S_IFSOCK,
    c_jffs2.DT_WHT: stat.S_IFWHT,
}



from __future__ import annotations

import logging
import os
import stat
import zlib
from bisect import bisect_right
from functools import cache, cached_property, lru_cache
from typing import TYPE_CHECKING, BinaryIO

from dissect.util import ts
from dissect.util.compression import lzo
from dissect.util.stream import AlignedStream, RunlistStream

from dissect.jffs.c_jffs2 import DT_MAP, JFFS2_MAGIC_NUMBERS, c_jffs2
from dissect.jffs.exceptions import (
    Error,
    FileNotFoundError,
    NotADirectoryError,
    NotASymlinkError,
)

if TYPE_CHECKING:
    from collections.abc import Iterator
    from datetime import datetime

log = logging.getLogger(__name__)
log.setLevel(os.getenv("DISSECT_LOG_JFFS", "CRITICAL"))


class JFFS2:
    """Journalling Flash File System, version 2 implementation.

    Currently supports the most common compression formats (``zlib``, ``lzo``, ``none`` and ``zero``).
    CRC checksums are currently ignored. Compression formats ``rtime``, ``rubinmips``, ``copy``
    and ``dynrubin`` are not supported.

    References:
        - https://sourceware.org/jffs2/
        - https://github.com/torvalds/linux/blob/master/fs/jffs2/readinode.c
    """

    def __init__(self, fh: BinaryIO):
        self.fh = fh

        # Dict of ino to list of tuples of (inode, offset)
        self._inodes: dict[int, list[tuple[c_jffs2.jffs2_raw_inode, int]]] = {}
        # Dict of pino to dict of dirent names to list of dirents
        self._dirents: dict[int, dict[bytes, list[DirEntry]]] = {}
        self._lost_found: list[list[DirEntry]] = []
        # Map of nlinks by inum
        self._nlinks_by_inum: dict[int, int] = {}

        if (node := c_jffs2.jffs2_unknown_node(fh)).magic not in JFFS2_MAGIC_NUMBERS:
            raise Error(f"Unknown JFFS2 magic: {node.magic:#x}")

        self.inode = lru_cache(4096)(self.inode)

        self.root = self.inode(inum=0x1, type=c_jffs2.DT_DIR, parent=None)

        self._scan()
        self._count_nlinks()
        self._garbage_collect()

    def inode(self, inum: int, type: int | None = None, parent: INode | None = None) -> INode:
        return INode(self, inum, type, parent)

    def get(self, path: str | int, node: INode | None = None) -> INode:
        if isinstance(path, int):
            return self.inode(path)

        parts = path.encode().split(b"/")
        node = node or self.root

        for part in parts:
            if not part:
                continue

            while node.type == stat.S_IFLNK:
                node = node.link_inode

            if dirent := self._dirents[node.inum].get(part):
                node = self.inode(dirent[-1].inum)
            else:
                raise FileNotFoundError(f"File not found: {path}")

        return node

    def _scan(self) -> None:
        """Scan the filesystem for inodes and dirents."""
        fh = self.fh
        pos = 0

        while True:
            fh.seek(pos)

            # Cheat a little and bypass cstruct for performance reasons
            buf = fh.read(4)
            if len(buf) != 4:
                break

            header = int.from_bytes(buf, "little")
            magic = header & 0xFFFF
            nodetype = header >> 16

            if magic not in JFFS2_MAGIC_NUMBERS:
                pos += 4
                continue

            if nodetype == c_jffs2.JFFS2_NODETYPE_DIRENT:
                self.fh.seek(pos)
                dirent = c_jffs2.jffs2_raw_dirent(self.fh)

                # If the dirent.ino is 0x00 this indicates that the dirent
                # and any associated data inodes are unlinked.
                if dirent.ino == 0:
                    orphaned = self._dirents[dirent.pino].pop(dirent.name)
                    self._lost_found.append(orphaned)
                else:
                    dirents = self._dirents.setdefault(dirent.pino, {}).setdefault(dirent.name, [])
                    entry = DirEntry(self, dirent)

                    # Only assure that the last entry is sorted correctly
                    if dirents and dirents[-1].version < entry.version:
                        dirents.append(entry)
                    else:
                        dirents.insert(0, entry)

                totlen = dirent.totlen

            elif nodetype == c_jffs2.JFFS2_NODETYPE_INODE:
                self.fh.seek(pos)
                inode = c_jffs2.jffs2_raw_inode(self.fh)

                inodes = self._inodes.setdefault(inode.ino, [])
                entry = (inode, pos + len(c_jffs2.jffs2_raw_inode))

                # Only assure that the last entry is sorted correctly
                if inodes and inodes[-1][0].version < inode.version:
                    inodes.append(entry)
                else:
                    inodes.insert(0, entry)

                totlen = inode.totlen

            else:
                log.warning("Unknown nodetype %#x, skipping", nodetype)
                try:
                    totlen = int.from_bytes(fh.read(4), "little")
                except EOFError:
                    break

            pos += (totlen + 3) & ~3

    def _count_nlinks(self) -> None:
        """Count the number of hardlinks for each inode.

        JFFS does not store nlink information in the inode itself, so we have to calculate it.
        """
        for direntries in self._dirents.values():
            for versions in direntries.values():
                if not versions:
                    continue

                last_version = versions[-1]
                if last_version.type == c_jffs2.DT_DIR:
                    # Root dir (inum == 1) gets three nlinks
                    # (see https://github.com/torvalds/linux/blob/6485cf5ea253d40d507cd71253c9568c5470cd27/fs/jffs2/fs.c#L311)
                    self._nlinks_by_inum[last_version.inum] = 3 if last_version.inum == 1 else 2

                    # Now update the parent entry, which might not have an associated nlink yet.
                    # The parent directory gets one nlink for each child directory.
                    base_nlinks = 3 if last_version.parent_inum == 1 else 2
                    self._nlinks_by_inum[last_version.parent_inum] = (
                        self._nlinks_by_inum.get(last_version.parent_inum, base_nlinks) + 1
                    )

                elif last_version.type == c_jffs2.DT_REG:
                    self._nlinks_by_inum[last_version.inum] = self._nlinks_by_inum.get(last_version.inum, 0) + 1
                elif last_version.type == c_jffs2.DT_LNK:
                    self._nlinks_by_inum[last_version.inum] = 1

    def _garbage_collect(self) -> None:
        """Collect all found orphaned files and put them in the lost+found folder."""
        if not self._lost_found:
            log.debug("No files to collect for lost+found folder.")
            return

        # Add lost+found folder to root
        trash_dirent = DirEntry(
            self,
            c_jffs2.jffs2_raw_dirent(
                pino=self.root.inum,
                version=1,
                ino=-1,
                type=c_jffs2.DT_DIR,
                name=b"lost+found",
            ),
        )
        self._dirents[self.root.inum][trash_dirent.entry.name] = [trash_dirent]
        self._dirents[-1] = {}

        # Add all orphaned files to the lost+found folder
        for dirents in self._lost_found:
            dirent = dirents[-1]
            dirent.name = f"{dirent.name}_ino_{dirent.inum}_pino_{dirent.parent_inum}_ver_{dirent.version}"
            dirent.parent_inum = -1
            self._dirents[-1][dirent.name.encode()] = dirents


class DirEntry:
    def __init__(self, fs: JFFS2, entry: c_jffs2.jffs2_raw_dirent):
        self.fs = fs
        self.entry = entry

        self.parent_inum = entry.pino
        self.version = entry.version
        self.inum = entry.ino
        self.type = entry.type

    @cached_property
    def name(self) -> str:
        return self.entry.name.decode(errors="surrogateescape")


class INode:
    def __init__(self, fs: JFFS2, inum: int, type: int | None = None, parent: INode | None = None):
        self.fs = fs
        self.inum = inum
        self._type = type
        self.parent = parent

        self.listdir = cache(self.listdir)

    def __repr__(self) -> str:
        return f"<INode inum={self.inum}>"

    @cached_property
    def inodes(self) -> list[tuple[c_jffs2.jffs2_raw_inode, int]]:
        # Root inode does not exist in jffs2 so we create a fake one (0x1),
        # -1 is reserved for the lost+found directory
        if self.inum in [0x1, -1]:
            return [
                (
                    c_jffs2.jffs2_raw_inode(
                        ino=self.inum,
                        version=0x1,
                        mode=0o40755,
                        uid=0x0,
                        gid=0x0,
                        atime=0x0,
                        mtime=0x0,
                        ctime=0x0,
                        csize=0x0,
                        dsize=0x0,
                    ),
                    0,
                )
            ]

        inodes = self.fs._inodes.get(self.inum)
        if not inodes:
            raise ValueError(f"INode with inum {self.inum} does not exist")

        return inodes

    @cached_property
    def inode(self) -> c_jffs2.jffs2_raw_inode:
        return self.inodes[-1][0]

    @cached_property
    def size(self) -> int:
        return self.inode.isize

    @cached_property
    def mode(self) -> int:
        return self.inode.mode

    @cached_property
    def type(self) -> int:
        return DT_MAP.get(self._type) or stat.S_IFMT(self.inode.mode)

    @cached_property
    def atime(self) -> datetime:
        return ts.from_unix(self.inode.atime)

    @cached_property
    def mtime(self) -> datetime:
        return ts.from_unix(self.inode.mtime)

    @cached_property
    def ctime(self) -> datetime:
        return ts.from_unix(self.inode.ctime)

    @cached_property
    def uid(self) -> int:
        return self.inode.uid

    @cached_property
    def gid(self) -> int:
        return self.inode.gid

    @cached_property
    def nlink(self) -> int:
        return self.fs._nlinks_by_inum.get(self.inum, 0)

    def is_dir(self) -> bool:
        return self.type == stat.S_IFDIR

    def is_file(self) -> bool:
        return self.type == stat.S_IFREG

    def is_symlink(self) -> bool:
        return self.type == stat.S_IFLNK

    @cached_property
    def link(self) -> str:
        if not self.is_symlink():
            raise NotASymlinkError(f"{self!r} is not a symlink")

        return self.open().read().decode(errors="surrogateescape")

    @cached_property
    def link_inode(self) -> INode:
        link = self.link
        return self.fs.get(self.link, self.parent if link.startswith("/") else None)

    def listdir(self) -> dict:
        return dict(self.iterdir())

    def iterdir(self) -> Iterator[tuple[str, INode]]:
        if not self.is_dir():
            raise NotADirectoryError(f"{self!r} is not a directory")

        for dirents in self.fs._dirents.get(self.inum, {}).values():
            dirent = dirents[-1]
            yield dirent.name, self.fs.inode(dirent.inum, dirent.type, parent=self)

    def open(self) -> RunlistStream:
        """Return data contained in all associated data inodes.

        Supports JFFS2_COMPR_NONE, JFFS2_COMPR_ZERO, JFFS2_COMPR_ZLIB and JFFS2_COMPR_LZO.

        Does not support JFFS2_COMPR_RTIME, JFFS2_COMPR_RUBINMIPS, JFFS2_COMPR_COPY and JFFS2_COMPR_DYNRUBIN.
        """
        return DataStream(self.fs, self.inum, self.size)


class DataStream(AlignedStream):
    """JFFS2 buffered stream that provides easy aligned reads.

    To read file contents, the log inodes are played back in
    version order, to recreate a map of where each range of
    data is located on the physical medium.
    """

    def __init__(self, fs: JFFS2, inum: int, size: int):
        self.fs = fs
        self.inum = inum

        runs = []
        run_offsets = []

        offset = 0
        for entry, pos in fs._inodes.get(inum, []):
            if offset != 0:
                run_offsets.append(offset)

            if offset < entry.offset:
                runs.append((None, None, entry.offset - offset, c_jffs2.JFFS2_COMPR_NONE))
                offset = entry.offset
                run_offsets.append(offset)

            runs.append((pos, entry.csize, entry.dsize, entry.compr))
            offset += entry.dsize

        if offset < size:
            runs.append((None, None, size - offset, c_jffs2.JFFS2_COMPR_NONE))
            run_offsets.append(offset)

        self._runlist = runs
        self._runlist_offsets = run_offsets

        super().__init__(size)

    def _read(self, offset: int, length: int) -> bytes:
        r = []

        run_idx = bisect_right(self._runlist_offsets, offset)
        runlist_len = len(self._runlist)
        size = self.size

        while length > 0:
            if run_idx >= runlist_len:
                # We somehow requested more data than we have runs for
                break

            # If run_idx == 0, we only have a single run
            run_pos = offset - (0 if run_idx == 0 else self._runlist_offsets[run_idx - 1])
            run_offset, run_data_size, run_size, run_compr = self._runlist[run_idx]
            run_remaining = run_size - run_pos

            # Sometimes the self.size is way larger than what we actually have runs for?
            # Stop reading if we reach a negative run_remaining
            if run_remaining < 0:
                break

            read_count = min(size - offset, min(run_remaining, length))

            # Sparse run
            if run_offset is None:
                r.append(b"\x00" * read_count)
            else:
                if run_compr in (c_jffs2.JFFS2_COMPR_NONE, c_jffs2.JFFS2_COMPR_ZERO):
                    self.fs.fh.seek(run_offset + run_pos)
                    r.append(self.fs.fh.read(read_count))
                else:
                    self.fs.fh.seek(run_offset)
                    buf = self.fs.fh.read(run_data_size)
                    if run_compr == c_jffs2.JFFS2_COMPR_ZLIB:
                        buf = zlib.decompress(buf)
                    elif run_compr == c_jffs2.JFFS2_COMPR_LZO:
                        buf = lzo.decompress(buf)
                    else:
                        raise NotImplementedError(f"Unsupported compression: {run_compr:#x}")

                    r.append(buf[run_pos : run_pos + read_count])

            offset += read_count
            length -= read_count
            run_idx += 1

        return b"".join(r)



class Error(Exception):
    pass


class FileNotFoundError(Error, FileNotFoundError):
    pass


class IsADirectoryError(Error, IsADirectoryError):
    pass


class NotADirectoryError(Error, NotADirectoryError):
    pass


class NotASymlinkError(Error):
    pass
